#/* Copyright 1998-2011 Research Foundation State University of New York */
#
#/* This file is part of QuB.                                            */
#
#/* QuB is free software; you can redistribute it and/or modify          */
#/* it under the terms of the GNU General Public License as published by */
#/* the Free Software Foundation, either version 3 of the License, or    */
#/* (at your option) any later version.                                  */
#
#/* QuB is distributed in the hope that it will be useful,               */
#/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
#/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
#/* GNU General Public License for more details.                         */
#
#/* You should have received a copy of the GNU General Public License,   */
#/* named LICENSE.txt, in the QuB program directory.  If not, see        */
#/* <http://www.gnu.org/licenses/>.                                      */

import os
import os.path
from itertools import *

from qubtree import *

import qubsettings
if not qubsettings.SettingsMgr:
    home = os.environ.get("HOME")
    if home:
        qubsettings.InitSettings('./qpr', os.path.join(home, ".qub", "qpr"))
    else:
        qubsettings.InitSettings('./qpr', './qprl')
SettingsMgr = qubsettings.SettingsMgr

import traceback
def tback():
    for line in traceback.format_stack():
        print line

from _qubfits import *

# _qubfits module methods:
#   bool      LoadPlugin("path/to/.so")
#   [name]    ListCurves()
#   Curve     CreateCurve(name)
#   [name]    ListFitters()
#   Fitter    CreateFitter(name)
#   Node      UseData(Node data, Node useSettings)
#   double    ChiSqrUncertainty(Curve, Node data, par_index, deltaPct=0.01)

# _qubfits.Curve properties:
#   componentCount  r/w
#   eqn             string, r/w if canSetEqn
#   canSetEqn       bool
#   paramCount
#   hasDeriv
#   msg             valid during "Curve Msg" and "Curve Err" events
#   state           qubtree.Node
# _qubfits.Curve methods:
#                   addEvent(name, f(name, Curve))
#                   delEvent(name, f(name, Curve))
#                   setExtraVars(Node data)
#   string          getParamName(par_index)
#   float           getParamVal(par_index)
#                   setParamVal(par_index, value)
#   bool            getParamConst(par_index)
#                   setParamConst(par_index, bool)
#                   initParams(Node data)
#                   evalAll(Node data)     into data['ff']
#                   evalOne(Node data)     into data['ff']
#                   evalDeriv(Node data)   into data['dPar'] (size paramcount)
#                   updateState(Node state)

# _qubfits.Fitter properties:
#   name
#   paramCount      fitter params (not curve params; e.g. max iterations)
#   iterations      valid during/after fitting
#   curve           valid after fitting
#   msg             valid during "Fitter Msg" and "Fitter Err" events
#   state           qubtree.Node
# _qubfits.Fitter methods:
#                   addEvent(name, f(name, Curve))
#                   delEvent(name, f(name, Curve))
#   string          getParamName(par_index)
#   float           getParamVal(par_index)
#                   setParamVal(par_index, value)
#   int             fit(Curve, Node data)
#                   stop()
#   float           getParamErr(par_index)
#                   updateState(Node state)

# direct-include qubtypes.py for no surprises/depends

import weakref

def we_printer(x):
    print x
    return x

class WeakEvent(object):
    """
    Usage:
       myEvent = WeakEvent() # a, b
       # note: keep a ref somewhere, even to self.whatever, or it will be unlinked
       callback1 = lambda a, b: dostuff(a, b)
       callback2 = lambda a, b: otherstuff(a, b)
       callback3 = self.callback3
       myEvent += callback1
       myEvent += callback2
       myEvent(a, b) # calls both callbacks
       myEvent -= callback1
       myEvent -= callback2
       myEvent(a, b) # no callbacks; does nothing
    """
       
    def __init__(self):
        self._listeners = []
        self.enabled = True
    def __call__(self, *args, **kw):
        dead = []
        if self.enabled:
            for l in self._listeners:
                listener = l()
                if listener:
                    listener(*args, **kw)
                else:
                    dead.append(l)
        for l in dead:
            print "dead"
            self._listeners.remove(l)
    def __iadd__(self, ln):
        try:
            l, n = ln
        except:
            l, n = ln, ""
        if not l in self._listeners:
            self._listeners.append(weakref.ref(l)) # , lambda l:we_printer("released "+n)))
        return self
    def __isub__(self, lsub):
        dead = []
        for l in self._listeners:
            listener = l()
            if (not listener) or (listener == lsub):
                dead.append(l)
        for l in dead:
            self._listeners.remove(l)
        if not len(dead):
            raise Exception('-=!~')
        return self

class DeathRattle(object):
    def __init__(self, rattle):
        self.rattle = rattle
    def __del__(self):
        print self.rattle




DISPLAY_DEFAULT = 0
DISPLAY_POINTS  = 1
DISPLAY_HIST    = 2

class Data(object):
    def __init__(self, prefix=''):
        self._source = NullNode()
        self._settings = None
        self._props = SettingsMgr[prefix+'FitData'].active.clone()
        self._initProps()
        self._absX = False
        self._displayStyle = DISPLAY_DEFAULT
        self.onChange = WeakEvent()
        self.onVarsChange = WeakEvent()
        self.xMin = 0.0
        self.xMax = 0.0
        self.yMin = 0.0
        self.yMax = 0.0
        self.yMinSel = 0.0
        self.yMaxSel = 0.0
        self.mins = {}
        self.maxs = {}
        #self.selFromX = 0.0
        #self.selToX = 0.0
        self.meanW = 1.0
        self.data = NullNode()
    source         = property(lambda s: s._source, lambda s,x: s._setSource(x))
    weights        = property(lambda s: s._getWeights(), lambda s,x: s._setWeights(x))
    settings       = property(lambda s: s._settings, lambda s,x: s._setSettings(x))
    xName          = property(lambda s: s._xName,
                              lambda s, x: s._setprop('_xName', "XName", x))
    yName          = property(lambda s: s._yName,
                              lambda s, x: s._setprop('_yName', "YName", x))
    absX           = property(lambda s: s._absX, lambda s,x: s._setAbsX(x))
    displayStyle   = property(lambda s: s._displayStyle, lambda s,x: s._setDisplayStyle(x))
    selFrom        = property(lambda s: s._selFrom,
                              lambda s, x: s._setprop('_selFrom', "SelFrom", max(0.0, x)))
    selTo          = property(lambda s: s._selTo,
                              lambda s, x: s._setprop('_selTo', "SelTo", min(100.0, x)))
    selFromS       = property(lambda s: s._getSelFromS(),
                              lambda s, x: s._setSelFromS(x))
    selToS         = property(lambda s: s._getSelToS(),
                              lambda s, x: s._setSelToS(x))
    doFilter       = property(lambda s: s._filter,
                              lambda s, x: s._setprop('_filter', "Filter", x))
    freq           = property(lambda s: s._freq,
                              lambda s, x: s._setprop('_freq', "Freq", x))
    resample       = property(lambda s: s._resample,
                              lambda s, x: s._setprop('_resample', "Resample", x))
    resampleStd    = property(lambda s: s._resampleStd,
                              lambda s, x: s._setprop('_resampleStd', "ResampleStd", x))
    def _setprop(self, field, name, val, updateSource=True):
        self.__dict__[field] = val
        self._props[name].data = val
        if updateSource:
            self._setSource(self.source, False) # update without vars-change
    def _initProps(self):
        p = self._props
        def initOneNode(node, val):
            if not node.data:
                node.data = val
            return node.data
        def initOne(name, val):
            return initOneNode(p[name], val)
        self._xName = str(initOne('XName', 'X'))
        self._yName = str(initOne('YName', 'Y'))
        self._wName = str(initOne('WName', ''))
        self._wInverse = bool(initOne('Winverse', 0)[0])
        self._selFrom = initOne('SelFrom', 0.0)[0]
        self._selTo = initOne('SelTo', 100.0)[0]
        self._filter = bool(initOne('Filter', 0)[0])
        self._freq = initOne('Freq', 1.0)[0]
        self._resample = bool(initOne('Resample', 0)[0])
        self._resampleStd = initOne('ResampleStd', 1.0)[0]
    def onPreSet(self, settings, updates):
        ndata = lambda name, defv: (lambda node: ((not node.isNull) and node.data[0]) or defv)(updates.find(name))
        sdata = lambda name, defv: (lambda node: ((not node.isNull) and str(node.data)) or defv)(updates.find(name))
        self._selFrom = ndata("SelFrom", self._selFrom)
        self._props["SelFrom"].data = self._selFrom
        self._selTo = ndata("SelTo", self._selTo)
        self._props["SelTo"].data = self._selTo
        self._filter = bool(ndata("Filter", self._filter))
        self._props["Filter"].data = self._filter
        self._freq = ndata("Freq", self._freq)
        self._props["Freq"].data = self._freq
        self._resample = bool(ndata("Resample", self._resample))
        self._props["Resample"].data = self._resample
        self._resampleStd = ndata("ResampleStd", self._resampleStd)
        self._props["ResampleStd"].data = self._resampleStd
        self._xName = sdata("XName", self._xName)
        self._props["XName"].data = self._xName
        self._yName = sdata("YName", self._yName)
        self._props["YName"].data = self._yName
        # allow blank WName:
        w = updates.find("WName")
        if not w.isNull:
            self._wName = str(w.data)
        self._props["WName"].data = self._wName
        self._wInverse = bool(ndata("Winverse", self._wInverse))
        self._props["Winverse"].data = self._wInverse
        # update
        self.source = self.source
    def _setSettings(self, x):
        if self._settings:
            bkup = self._props.clone()
            if x:
                self._settings.setProperties(x.active) # calls onPreSet to override old with new settings
            self._settings.active = bkup               # disentangle from old settings, unmodified
            self._settings.onSet = None
            self._settings.saveActive()
        elif x:
            self.onPreSet(x, x.active)
        self._settings = x
        if x:
            x.active = self._props
            self.hOnPreSet = self.onPreSet
            x.onSet = self.hOnPreSet
    def _getWeights(self):
        return (self._wName, self._wInverse)
    def _setWeights(self, x):
        self._wName = x[0]
        self._props["WName"].data = x[0]
        self._wInverse = x[1]
        self._props["Winverse"].data = x[1]
        self.source = self.source
    def _setAbsX(self, x):
        self._absX = x
        self._source["AbsX"].data = x
        self.source = self.source
    def _setDisplayStyle(self, x):
        self._displayStyle = x
        self._source["DisplayStyle"].data = x
        self.source = self.source
    def _reWeight(self):
        if self._source['n'].data.count:
            n = self._source["n"].data[0]
        else:
            n = 0
        if n == 0:
            self.meanW = 1.0
            return
        ww = self._source["ww"].data
        if ww.count == 0:
            ww.setup(QTR_TYPE_DOUBLE, n, 1)
        elif ww.count < n:
            ww.resize(n)
        ws = None
        if self._wName:
            for node in children(self._source["Series"]):
                if node.name == self._wName:
                    ws = node.data[:n]
                    break
        if not ws:
            ws = [1.0] * n
        if self._wInverse:
            ws = [(x and (1.0/x)) or 1.0 for x in ws]
        wsum = 0.0
        for i,w in enumerate(ws):
            ww[i] = w
            wsum += w
        self.meanW = wsum / n
    def _reMeasure(self):
        if (not self._source['n'].data) or (0 == self._source['n'].data[0]):
            return
        n = self._source['n'].data[0]
        for series in children(self._source['Series']):
            ss = series.data[:n]
            nm = series.name
            self.mins[nm] = min(ss)
            self.maxs[nm] = max(ss)
        self.xMin = self.mins[self.xName]
        self.xMax = self.maxs[self.xName]
        self.yMin = self.mins[self.yName]
        self.yMax = self.maxs[self.yName]
        n = self.data['n'].data[0]
        self.yMinSel = min(self.data["Series"][self._yName].data[:n])
        self.yMaxSel = max(self.data["Series"][self._yName].data[:n])
    def _setSource(self, x, varsChange=True):
        self._source = x
        if x['Series'].find(self._xName).data:
            self._source["XName"].data = self._xName
        else:
            self._setprop('_xName', "XName", str(x['XName'].data), False)
        if x['Series'].find(self._yName).data:
            self._source["YName"].data = self._yName
        else:
            self._setprop('_yName', "YName", str(x['YName'].data), False)
        if self._source.find('AbsX').data.count:
            self._absX = bool(self._source["AbsX"].data[0])
        else:
            self._source['AbsX'].data = self._absX
        if self._source['DisplayStyle'].data.count:
            self._displayStyle = self._source["DisplayStyle"].data[0]
        else:
            self._source['DisplayStyle'].data = self._displayStyle
        self._reWeight()
        if self._source:
            self.data = UseData(self._source, self._props)
        else:
            self.data = NullNode()
        self._reMeasure()
        if varsChange:
            self.onVarsChange(self)
        self.onChange(self)
    def _getSelFromS(self):
        if self.xMin != self.xMax:
            return self.xMin + (self.selFrom / 100.0) * (self.xMax - self.xMin)
        else:
            return self.xMin
    def _setSelFromS(self, x):
        if self.xMin != self.xMax:
            self.selFrom = 100.0 * (x - self.xMin) / (self.xMax - self.xMin)
    def _getSelToS(self):
        if self.xMin != self.xMax:
            return self.xMin + (self.selTo / 100.0) * (self.xMax - self.xMin)
        else:
            return self.xMin
    def _setSelToS(self, x):
        if self.xMin != self.xMax:
            self.selTo = 100.0 * (x - self.xMin) / (self.xMax - self.xMin)


class FitContext(object):
    def __init__(self, prefix=''):
        self._prefix = prefix
        self.onCurveChange = WeakEvent()
        self.onCurveCompCount = WeakEvent()
        self.onCurveEqn = WeakEvent()
        self.onCurveParams = WeakEvent()
        self.onCurveParamVal = WeakEvent()
        self.onCurveParamConst = WeakEvent()
        self.onCurveMsg = WeakEvent()
        self.onCurveErr = WeakEvent()
        self.onFitterChange = WeakEvent()
        self.onFitterParamVal = WeakEvent()
        self.onFitterIter = WeakEvent()
        self.onFitterMsg = WeakEvent()
        self.onFitterErr = WeakEvent()
        self.onFitterDone = WeakEvent()
        self.onDataChange = WeakEvent()
        self.onVarsChange = WeakEvent()
        self.onErrorsChange = WeakEvent()
        self.hOnCurveCompCount = self._onCurveCompCount
        self.hOnCurveEqn = self._onCurveEqn
        self.hOnCurveParams = self._onCurveParams
        self.hOnCurveParamVal = self._onCurveParamVal
        self.hOnCurveParamConst = self._onCurveParamConst
        self.hOnCurveMsg = self._onCurveMsg
        self.hOnCurveErr = self._onCurveErr
        self.hOnFitterParamVal = self._onFitterParamVal
        self.hOnFitterIter = self._onFitterIter
        self.hOnFitterMsg = self._onFitterMsg
        self.hOnFitterErr = self._onFitterErr
        self.errors = []
        self.residual = 0.0
        self._fitting = False
        self._lastIter = -1
        self.data = Data()
        self.data.settings = SettingsMgr[prefix+'FitData']
        self.hOnDataChange = self._onDataChange
        self.data.onChange += self.hOnDataChange, "onDataChange"
        self.hOnVarsChange = self._onVarsChange
        self.data.onVarsChange += self.hOnVarsChange, "onVarsChange"
        self._curve = None
        self._fitter = None
        self._initCurve()
        self._initFitter()
    curve = property(lambda self: self._curve, lambda self, x: self._setCurve(x))
    fitter = property(lambda self: self._fitter, lambda self, x: self._setFitter(x))
    def _initCurve(self):
        curveSet = SettingsMgr[self._prefix + 'Curve']
        details = curveSet.active.clone()
        curveList = ListCurves()
        restore = SettingsMgr[self._prefix + 'Curve Restore'].active
        if str(curveSet.active['Name'].data) in curveList:
            self.curve = CreateCurve(str(curveSet.active['Name'].data))
        else:
            self.curve = CreateCurve(curveList[0])
        self.curve.updateState(details)
        self.hOnCurvePreSet = self._onCurvePreSet
        curveSet.onSet = self.hOnCurvePreSet
    def _initFitter(self):
        fitterSet = SettingsMgr[self._prefix + 'Fitter']
        details = fitterSet.active.clone()
        fitterList = ListFitters()
        restore = SettingsMgr[self._prefix + 'Fitter Restore'].active
        if str(fitterSet.active['Name'].data) in fitterList:
            self.fitter = CreateFitter(str(fitterSet.active['Name'].data))
        else:
            self.fitter = CreateFitter(fitterList[0])
        self.fitter.updateState(details)
        self.hOnFitterPreSet = self._onFitterPreSet
        fitterSet.onSet = self.hOnFitterPreSet
    def _onCurvePreSet(self, settings, updates):
        upName = str(updates['Name'].data)
        if upname and (self.curve.name != upName):
            self.curve = CreateCurve(upName)
        self.curve.updateState(updates)
    def _onFitterPreSet(self, settings, updates):
        upName = str(updates['Name'].data)
        if upname and (self.fitter.name != upName):
            self.fitter = CreateFitter(upName)
        self.fitter.updateState(updates)
    def _setCurve(self, curve):
        restore = SettingsMgr[self._prefix + 'Curve Restore'].active
        if self._curve:
            reCurve = restore.find(self._curve.name)
            if not reCurve.isNull:
                restore.remove(reCurve)
            reCurve = self._curve.state.clone()
            reCurve.name = str(reCurve['Name'].data)
            restore.insert(reCurve)
            self._curve.delEvent('Curve CompCount', self.hOnCurveCompCount)
            self._curve.delEvent('Curve Eqn', self.hOnCurveEqn)
            self._curve.delEvent('Curve Params', self.hOnCurveParams)
            self._curve.delEvent('Curve ParamVal', self.hOnCurveParamVal)
            self._curve.delEvent('Curve ParamConst', self.hOnCurveParamConst)
            self._curve.delEvent('Curve Msg', self.hOnCurveMsg)
            self._curve.delEvent('Curve Err', self.hOnCurveErr)
        self._curve = curve
        if curve:
            curve.setExtraVars(self.data.data)
            SettingsMgr[self._prefix+'Curve'].active = curve.state
            self._curve.addEvent('Curve CompCount', self.hOnCurveCompCount)
            self._curve.addEvent('Curve Eqn', self.hOnCurveEqn)
            self._curve.addEvent('Curve Params', self.hOnCurveParams)
            self._curve.addEvent('Curve ParamVal', self.hOnCurveParamVal)
            self._curve.addEvent('Curve ParamConst', self.hOnCurveParamConst)
            self._curve.addEvent('Curve Msg', self.hOnCurveMsg)
            self._curve.addEvent('Curve Err', self.hOnCurveErr)
            reCurve = restore.find(curve.name)
            if not reCurve.isNull:
                curve.updateState(reCurve)
        self._reCalc()
        self.onCurveChange(curve)
    def _setFitter(self, fitter):
        restore = SettingsMgr[self._prefix+'Fitter Restore'].active
        if self._fitter:
            reFit = restore.find(self._fitter.name)
            if not reFit.isNull:
                restore.remove(reFit)
            reFit = self._fitter.state.clone()
            reFit.name = str(reFit['Name'].data)
            restore.insert(reFit)
            self._fitter.delEvent('Fitter ParamVal', self.hOnFitterParamVal)
            self._fitter.delEvent('Fitter Iter', self.hOnFitterIter)
            self._fitter.delEvent('Fitter Msg', self.hOnFitterMsg)
            self._fitter.delEvent('Fitter Err', self.hOnFitterErr)
        self._fitter = fitter
        if fitter:
            SettingsMgr[self._prefix+'Fitter'].active = fitter.state
            self._fitter.addEvent('Fitter ParamVal', self.hOnFitterParamVal)
            self._fitter.addEvent('Fitter Iter', self.hOnFitterIter)
            self._fitter.addEvent('Fitter Msg', self.hOnFitterMsg)
            self._fitter.addEvent('Fitter Err', self.hOnFitterErr)
            reFit = restore.find(fitter.name)
            if not reFit.isNull:
                fitter.updateState(reFit)
        self.onFitterChange(fitter)
    def fit(self):
        try:
            self._lastIter = 0
            self._fitting = True
            self.fitter.fit(self.curve, self.data.data)
            self.errors = [self.fitter.getParamErr(i) for i in xrange(self.curve.paramCount)]
            self.onErrorsChange(self)
        except:
            pass # ?
        self._fitting = False
        self.onFitterDone(self.fitter)
    def stop(self):
        self.fitter.stop()
    def _reCalc(self):
        if not self.curve:
            return
        self.errors = [0.0] * self.curve.paramCount
        self.residual = 0.0
        if not self.curve: return
        if self.data.data.find('n').data:
            n = self.data.data['n'].data[0]
        else:
            n = 0
        if n <= 0: return
        self._fitting = True
        self.curve.evalAll(self.data.data)
        self.errors = [ChiSqrUncertainty(self.curve, self.data.data, i) for i in xrange(self.curve.paramCount)]
        self.onErrorsChange(self)
        self._fitting = False
        ff = self.data.data['ff'].data[:n]
        yy = self.data.data['Series'][str(self.data.data['YName'].data)].data[:n]
        self.residual = reduce( lambda a,b: a+b, imap(lambda x,y: (x-y)**2, ff, yy) )
    def _onDataChange(self, data):
        self._reCalc()
        self.onDataChange(data)
    def _onVarsChange(self, data):
        if self.curve:
            self.curve.setExtraVars(self.data.data)
        self.onVarsChange(data)
    def _onCurveCompCount(self, name, curve):
        self.onCurveCompCount(curve, curve.componentCount)
    def _onCurveEqn(self, name, curve):
        self.onCurveEqn(curve, curve.eqn)
    def _onCurveParams(self, name, curve):
        self._reCalc()
        self.onCurveParams(curve)
    def _onCurveParamVal(self, name, curve):
        if not self._fitting:
            self._reCalc()
        self.onCurveParamVal(curve)
    def _onCurveParamConst(self, name, curve):
        self.onCurveParamConst(curve)
    def _onCurveMsg(self, name, curve):
        self.onCurveMsg(curve, curve.msg)
    def _onCurveErr(self, name, curve):
        self.onCurveErr(curve, curve.msg)
    def _onFitterParamVal(self, name, fitter):
        self.onFitterParamVal(fitter)
    def _onFitterIter(self, name, fitter):
        its = fitter.iterations
        if its != self._lastIter:
            self._lastIter = its
            self._reCalc()
            self.onFitterIter(fitter, fitter.iterations)
    def _onFitterMsg(self, name, fitter):
        self.onFitterMsg(fitter, fitter.msg)
    def _onFitterErr(self, name, fitter):
        self.onFitterErr(fitter, fitter.msg)

def LoadPlugins(path):
    for root, dirs, files in os.walk(path):
        for name in files:
            try:
                LoadPlugin(os.path.join(root, name))
            except Exception, e:
                print e
        dirs[:] = [] #don't recurse into sub-directories


if __name__ == '__main__':
    #LoadPlugins("/home/cnicolai/lib/qubfits")
    LoadPlugins("plugins")
    
    def onMsg(obj, msg):
        print obj,":",msg
        
    c = FitContext()
    c.onCurveMsg += onMsg
    c.onCurveErr += onMsg
    c.onFitterMsg += onMsg
    c.onFitterErr += onMsg

    # example of program-generated data
    #data = qubtree.Node()
    #data['n'].data = 5 # number of data points
    #data['Series']['X'].data = (0.0, 0.1, 0.2, 0.3, 0.4)
    #data['XName'].data = 'X'
    #data['Series']['Y'].data = (5.0, 4.0, 3.0, 2.0, 1.0)
    #data['YName'].data = 'Y'
    #data['Series']['W'].data = (1.0, 1.0, 1.0, 1.0, 1.0) # bigger weight more important
    #c.data.source = data
    #c.data.wName = 'W'

    # actually using a data file from disk
    c.data.source = Open('data.qtr')
    
    print c.data.source
    
    print "Installed:",c.curve
    print "Installed:",c.fitter
    print c.curve.state
    print "Residual:",c.residual
    print "Errors:",c.errors

    c.curve = CreateCurve("Custom")
    c.curve.setParamVal(0, 1.8)
    c.curve.setParamVal(1, 32.0)
    print "Residual:",c.residual
    print "Errors:",c.errors
    
    c.fitter = CreateFitter("Levenburg-Marquardt")
    c.fit()
    print 'iterations:', c.fitter.iterations
    print c.curve.state
    print c.data.data
    print "Residual:",c.residual
    print "Errors:",c.errors
    
    qubsettings.SettingsMgr.saveAll()
